#pragma once


// STD/STL
#include <functional>
#include <unordered_map>
#include <vector>

// RobotAPI
#include <RobotAPI/interface/armem/server/ReadingMemoryInterface.h>
#include <RobotAPI/interface/armem/server/StoringMemoryInterface.h>
#include <RobotAPI/interface/armem/client/MemoryListenerInterface.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>
#include <RobotAPI/libraries/armem/core/workingmemory/Memory.h>
#include <RobotAPI/libraries/armem/core/longtermmemory/Memory.h>

#include "Query.h"


namespace armarx::armem::client
{

    /**
     * @brief Reads data from a memory server.
     */
    class Reader
    {

        using callback = std::function<void(const MemoryID& subscriptionID, const std::vector<MemoryID>& updatedSnapshotIDs)>;
        using callback_updated_only = std::function<void(const std::vector<MemoryID>& updatedSnapshotIDs)>;

        template <class CalleeT>
        using member_callback = void(CalleeT::*)(const MemoryID& subscriptionID, const std::vector<MemoryID>& updatedSnapshotIDs);
        template <class CalleeT>
        using member_callback_updated_only = void(CalleeT::*)(const std::vector<MemoryID>& updatedSnapshotIDs);


    public:

        /**
         * @brief Construct a memory reader.
         * @param memory The memory proxy.
         */
        Reader(server::ReadingMemoryInterfacePrx memory = nullptr);

        void setReadingMemory(server::ReadingMemoryInterfacePrx memory);

        QueryResult query(const QueryInput& input) const;
        armem::query::data::Result query(const armem::query::data::Input& input) const;

        QueryResult query(armem::query::data::MemoryQueryPtr query, DataMode dataMode = DataMode::WithData) const;
        QueryResult query(const armem::query::data::MemoryQuerySeq& queries, DataMode dataMode = DataMode::WithData) const;


        QueryResult getAll(DataMode dataMode = DataMode::WithData) const;
        QueryResult getLatestSnapshots(DataMode dataMode = DataMode::WithData) const;

        //data::StoreResult readAndStore(data::StoreInputSeq& input);
        data::StoreResult readAndStore(const data::StoreInput& input) const;

        void subscribe(const MemoryID& subscriptionID, callback callback);
        void subscribe(const MemoryID& subscriptionID, callback_updated_only callback);
        /**
         * Subscribe with a class member function:
         * @code
         * reader.subscribe(entityID, this, &This::myCallback);
         * @endcode
         */
        template <class CalleeT>
        void subscribe(const MemoryID& subscriptionID, CalleeT* callee, member_callback<CalleeT> callback)
        {
            auto cb = [callee, callback](const MemoryID & subscriptionID, const std::vector<MemoryID>& updatedSnapshotIDs)
            {
                (callee->*callback)(subscriptionID, updatedSnapshotIDs);
            };
            subscribe(subscriptionID, cb);
        }
        template <class CalleeT>
        void subscribe(const MemoryID& subscriptionID, CalleeT* callee, member_callback_updated_only<CalleeT> callback)
        {
            auto cb = [callee, callback](const MemoryID&, const std::vector<MemoryID>& updatedSnapshotIDs)
            {
                (callee->*callback)(updatedSnapshotIDs);
            };
            subscribe(subscriptionID, cb);
        }
        /// Function handling updates from the MemoryListener ice topic.
        void updated(const std::vector<MemoryID>& updatedIDs) const;


        inline operator bool() const
        {
            return bool(memoryPrx);
        }

    public:

        server::ReadingMemoryInterfacePrx memoryPrx;

    private:

        std::unordered_map<MemoryID, std::vector<callback>> callbacks;

    };

}
