#include "MappingDataReader.h"

// STD / STL
#include <cstring>
#include <vector>
#include <algorithm>
#include <map>
#include <optional>
#include <ostream>
#include <type_traits>
#include <utility>

// ICE
#include <IceUtil/Time.h>
#include <IceUtil/Handle.h>

// Simox
#include <SimoxUtility/algorithm/get_map_keys_values.h>

// ArmarXCore
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/logging/LogSender.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

// RobotAPI Interfaces
#include <RobotAPI/interface/units/LaserScannerUnit.h>
#include <RobotAPI/interface/armem/mns/MemoryNameSystemInterface.h>
#include <RobotAPI/interface/armem/server/ReadingMemoryInterface.h>

// RobotAPI Aron
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/complex/NDArray.h>

// RobotAPI Armem
#include <RobotAPI/libraries/armem/client/Query.h>
#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/selectors.h>

#include <RobotAPI/libraries/armem/core/workingmemory/Memory.h>
#include <RobotAPI/libraries/armem/core/workingmemory/CoreSegment.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ProviderSegment.h>
#include <RobotAPI/libraries/armem/core/workingmemory/Entity.h>
#include <RobotAPI/libraries/armem/core/workingmemory/EntityInstance.h>

#include <RobotAPI/libraries/armem/util/util.h>

#include <RobotAPI/libraries/armem_robot_mapping/aron/LaserScan.aron.generated.h>
#include <RobotAPI/libraries/armem_robot_mapping/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_mapping/types.h>

namespace armarx::armem
{

    MappingDataReader::MappingDataReader(armem::ClientReaderComponentPluginUser& memoryClient)
        : memoryClient(memoryClient) {}

    MappingDataReader::~MappingDataReader() = default;

    void MappingDataReader::registerPropertyDefinitions(
        armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "TransformReader: registerPropertyDefinitions";
        registerPropertyDefinitions(def);

        const std::string prefix = propertyPrefix;

        def->optional(properties.mappingMemoryName, prefix + "MappingMemoryName",
                      "Name of the mapping memory core segment to use.");

        def->optional(properties.memoryName, prefix + "MemoryName");
    }

    void MappingDataReader::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "TransformReader: Waiting for memory '"
                         << properties.memoryName << "' ...";
        auto result = memoryClient.useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "TransformReader: Connected to memory '"
                         << properties.memoryName;

        memoryReader.setReadingMemory(result.proxy);
    }

    armem::client::query::Builder
    MappingDataReader::buildQuery(const Query& query) const
    {
        armem::client::query::Builder qb;

        ARMARX_INFO << "Query for agent: " << query.agent
                    << " memory name: " << properties.mappingMemoryName;

        if (query.sensorList.empty()) // all sensors
        {
            // clang-format off
            qb
            .coreSegments().withName(properties.mappingMemoryName)
            .providerSegments().withName(query.agent)
            .entities().all()
            .snapshots().timeRange(query.timeRange.min, query.timeRange.max);
            // clang-format on
        }
        else
        {
            // clang-format off
            qb
            .coreSegments().withName(properties.mappingMemoryName)
            .providerSegments().withName(query.agent)
            .entities().withNames(query.sensorList)
            .snapshots().timeRange(query.timeRange.min, query.timeRange.max);
            // clang-format on
        }

        return qb;

    }

    std::vector<LaserScanStamped> asLaserScans(const std::map<std::string, wm::Entity>& entities)
    {
        std::vector<LaserScanStamped> outV;

        if (entities.empty())
        {
            ARMARX_WARNING << "No entities!";
        }

        const auto convert = [](const arondto::LaserScanStamped & aronLaserScanStamped, const wm::EntityInstance & ei) -> LaserScanStamped
        {
            LaserScanStamped laserScanStamped;
            fromAron(aronLaserScanStamped, laserScanStamped);

            const auto ndArrayNavigator = aron::datanavigator::NDArrayNavigator::DynamicCast(ei.data()->getElement("scan"));

            ARMARX_CHECK_NOT_NULL(ndArrayNavigator);

            laserScanStamped.data = fromAron<LaserScanStep>(ndArrayNavigator);
            ARMARX_IMPORTANT << "4";


            return laserScanStamped;

        };

        // loop over all entities and their snapshots
        for (const auto &[s, entity] : entities)
        {
            if (entity.empty())
            {
                ARMARX_WARNING << "Empty history for " << s;
            }

            ARMARX_DEBUG << "History size: " << entity.size();

            for (const auto &[ss, entitySnapshot] : entity)
            {
                for (const auto& entityInstance : entitySnapshot.instances())
                {
                    const auto o = tryCast<arondto::LaserScanStamped>(entityInstance);

                    if (o)
                    {
                        outV.push_back(convert(*o, entityInstance));
                    }
                }
            }
        }

        return outV;
    }

    MappingDataReader::Result
    MappingDataReader::queryData(const Query& query) const
    {
        const auto qb = buildQuery(query);

        ARMARX_IMPORTANT << "[MappingDataReader] query ... ";

        const armem::client::QueryResult qResult =
            memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "[MappingDataReader] result: " << qResult;

        if (not qResult.success)
        {
            ARMARX_WARNING << "Failed to query data from memory: "
                           << qResult.errorMessage;
            return {.laserScans = {},
                    .sensors = {},
                    .status = Result::Status::Error,
                    .errorMessage = qResult.errorMessage};
        }

        // now create result from memory
        const auto& entities =
            qResult.memory.getCoreSegment(properties.mappingMemoryName)
            .getProviderSegment(query.agent)
            .entities();

        const auto laserScans = asLaserScans(entities);
        const auto sensors = simox::alg::get_keys(entities);

        return {.laserScans = laserScans,
                .sensors = sensors,
                .status = Result::Status::Success,
                .errorMessage = ""};
    }

} // namespace armarx::armem
