/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>
#include <numeric>

// Eigen
#include <Eigen/Geometry>
#include <Eigen/Core>

// ArmarX
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/navigator/data/complex/NDArray.h>


namespace armarx::aron::converter
{
    class AronEigenConverter
    {
        AronEigenConverter() = delete;

    public:
        // TODO: Add inplace methods

        static Eigen::Quaternion<float> ConvertToQuaternionf(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Quaternion<float> ConvertToQuaternionf(const datanavigator::NDArrayNavigator&);
        static Eigen::Quaternion<double> ConvertToQuaterniond(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Quaternion<double> ConvertToQuaterniond(const datanavigator::NDArrayNavigator&);
        static Eigen::Vector3f ConvertToVector3f(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Vector3f ConvertToVector3f(const datanavigator::NDArrayNavigator&);
        static Eigen::Vector3d ConvertToVector3d(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Vector3d ConvertToVector3d(const datanavigator::NDArrayNavigator&);
        static Eigen::Matrix4f ConvertToMatrix4f(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Matrix4f ConvertToMatrix4f(const datanavigator::NDArrayNavigator&);
        static Eigen::Matrix4d ConvertToMatrix4d(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Matrix4d ConvertToMatrix4d(const datanavigator::NDArrayNavigator&);


        template<typename T>
        static Eigen::Quaternion<T> ConvertToQuaternion(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToQuaternion<T>(*nav);
        }

        template<typename T>
        static Eigen::Quaternion<T> ConvertToQuaternion(const datanavigator::NDArrayNavigator& nav)
        {
            checkDimensions(nav, {1, 4, sizeof(T)}, "ConvertToQuaternion");
            auto dims = nav.getDimensions();

            Eigen::Quaternion<T> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.coeffs().data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }


        template<typename T, int Size>
        static Eigen::Matrix<T, Size, 1> ConvertToVector(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToVector<T, Size>(*nav);
        }

        template<typename T, int Size>
        static Eigen::Matrix<T, Size, 1> ConvertToVector(const datanavigator::NDArrayNavigator& nav)
        {
            checkDimensions(nav, {Size, 1, sizeof(T)}, "ConvertToVector");
            auto dims = nav.getDimensions();

            Eigen::Matrix<T, Size, 1> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }


        template<typename T, int Rows, int Cols>
        static Eigen::Matrix<T, Rows, Cols> ConvertToMatrix(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToMatrix<T, Rows, Cols>(*nav);
        }

        template<typename T, int Rows, int Cols>
        static Eigen::Matrix<T, Rows, Cols> ConvertToMatrix(const datanavigator::NDArrayNavigator& nav)
        {
            checkDimensions(nav, {Rows, Cols, sizeof(T)}, "ConvertToMatrix");
            auto dims = nav.getDimensions();

            Eigen::Matrix<T, Rows, Cols> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }

    private:

        /**
         * @throw `error::AronException` If `nav`'s dimensions do not match `expected`.
         */
        static void checkDimensions(const datanavigator::NDArrayNavigator& nav, const std::vector<int>& expected,
                                    const std::string& method, const std::string& caller = "AronEigenConverter");

    };

}
