/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

/**
 * This file contains the general structure of all accepted AronTypes.
 *
 * In general we distinguish we distinguish between container(complex) and primitive types.
 * Complex types include Objects, Lists, Dicts, Pairs and Blobs.
 * Pimitive types on the other hand include int, long, float, double, string and bools.
 *
 * A primitive always has only one member containing the value of the aron. A container type can have multiple members.
 *
 * Note that every AronData must have a corresponding AronAbstractType (both declared in aron.ice).
 * The AronData is a copy of cpp values while the AronAbstractType describes the type structure.
 *
 * In the following we descripe the types in more detail:
 * AronData:
 * |
 * |->Container:
 * |  |
 * |  |->AronObject:
 * |  |    This container represents cpp classes. It contains a std::map mapping from a membername (std::string) to AronData.
 * |  |    It is a recursive structure that again can contain other complex or primitive types.
 * |  |
 * |  |->AronList:
 * This container represents a std::vector<T> where T can be any datatype convertable to Aron.
 * TODO: complete doc
 */


#include <string>

namespace armarx
{
    namespace aron
    {
        // This simplifies code generation.
        // It may be removed in favor of another solution in the future.
        using string = std::string;
    }
}


/************************************
 * TYPES ****************************
 ***********************************/

// All containertypes go here
#define HANDLE_CONTAINER_TYPES \
    RUN_ARON_MACRO(Object, object, OBJECT) \
    RUN_ARON_MACRO(List, list, LIST) \
    RUN_ARON_MACRO(Dict, dict, DICT) \
    RUN_ARON_MACRO(Pair, pair, PAIR) \
    RUN_ARON_MACRO(Tuple, tuple, TUPLE)

#define HANDLE_DICT_SERIALIZER_TYPES \
    RUN_ARON_MACRO(Object, object, OBJECT) \
    RUN_ARON_MACRO(Dict, dict, DICT)

#define HANDLE_LIST_SERIALIZER_TYPES \
    RUN_ARON_MACRO(List, list, LIST) \
    RUN_ARON_MACRO(Tuple, tuple, TUPLE) \
    RUN_ARON_MACRO(Pair, pair, PAIR)

// All complex types go here
#define HANDLE_NDARRAY_TYPES \
    RUN_ARON_MACRO(EigenMatrix, eigenmatrix, EIGEN_MATRIX) \
    RUN_ARON_MACRO(EigenQuaternion, eigenquaternion, EIGEN_QUATERNION) \
    RUN_ARON_MACRO(IVTCByteImage, ivtcbyteimage, IVT_CBYTE_IMAGE) \
    RUN_ARON_MACRO(OpenCVMat, opencvmat, OPENCV_MAT) \
    RUN_ARON_MACRO(PCLPointCloud, pclpointcloud, PCL_POINTCLOUD) \
    RUN_ARON_MACRO(Position, position, POSITION) \
    RUN_ARON_MACRO(Orientation, orientation, ORIENTATION) \
    RUN_ARON_MACRO(Pose, pose, POSE)

// All enum types (they differ from primitive types since they have a list of accepted values and names)
// TODO: Think about extending enums to StringEnums, FloatEnumns etc.
#define HANDLE_ENUM_TYPES \
    RUN_ARON_MACRO(IntEnum, intenum, INT_ENUM)

// All primitive types
#define HANDLE_PRIMITIVE_TYPES \
    RUN_ARON_MACRO(Int, int, INT) \
    RUN_ARON_MACRO(Long, long, LONG) \
    RUN_ARON_MACRO(Float, float, FLOAT) \
    RUN_ARON_MACRO(Double, double, DOUBLE) \
    RUN_ARON_MACRO(String, string, STRING) \
    RUN_ARON_MACRO(Bool, bool, BOOL) \
    RUN_ARON_MACRO(Time, time, TIME)

// All combined
#define HANDLE_ALL_ARON_TYPES \
    HANDLE_CONTAINER_TYPES \
    HANDLE_NDARRAY_TYPES \
    HANDLE_ENUM_TYPES \
    HANDLE_PRIMITIVE_TYPES

/************************************
 * DATA *****************************
 ***********************************/

// All containerdata go here
#define HANDLE_CONTAINER_DATA \
    RUN_ARON_MACRO(List, list, LIST) \
    RUN_ARON_MACRO(Dict, dict, DICT)

#define HANDLE_COMPLEX_DATA \
    RUN_ARON_MACRO(NDArray, ndarray, NDARRAY)

// the lowercase name must be euqal to the cpp name
#define HANDLE_PRIMITIVE_DATA \
    RUN_ARON_MACRO(Int, int, INT) \
    RUN_ARON_MACRO(Long, long, LONG) \
    RUN_ARON_MACRO(Float, float, FLOAT) \
    RUN_ARON_MACRO(Double, double, DOUBLE) \
    RUN_ARON_MACRO(String, string, STRING) \
    RUN_ARON_MACRO(Bool, bool, BOOL)

// if you only need the "real" primitives
#define HANDLE_PRIMITIVE_DATA_EXCEPT_STRING \
    RUN_ARON_MACRO(Int, int, INT) \
    RUN_ARON_MACRO(Long, long, LONG) \
    RUN_ARON_MACRO(Float, float, FLOAT) \
    RUN_ARON_MACRO(Double, double, DOUBLE) \
    RUN_ARON_MACRO(Bool, bool, BOOL)

// All three combined
#define HANDLE_ALL_ARON_DATA \
    HANDLE_CONTAINER_DATA \
    HANDLE_COMPLEX_DATA \
    HANDLE_PRIMITIVE_DATA

/************************************
 * What type corresponds to which data
 * (t,t,T,d,d,D) => t in Type, d in Data
 ***********************************/
#define HANDLE_CONTAINER_CORRESPONDING \
    RUN_ARON_MACRO(List, list, LIST, List, list, LIST) \
    RUN_ARON_MACRO(Pair, pair, PAIR, List, list, LIST) \
    RUN_ARON_MACRO(Tuple, tuple, TUPLE, List, list, LIST) \
    RUN_ARON_MACRO(Object, object, OBJECT, Dict, dict, DICT) \
    RUN_ARON_MACRO(Dict, dict, DICT, Dict, dict, DICT)

#define HANDLE_COMPLEX_CORRESPONDING \
    RUN_ARON_MACRO(EigenMatrix, eigenmatrix, EIGEN_MATRIX, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(EigenQuaternion, eigenmatrix, EIGEN_MATRIX, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(IVTCByteImage, ivtcbyteimage, IVT_CBYTE_IMAGE, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(OpenCVMat, opencvmat, OPENCV_MAT, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(PCLPointCloud, pclpointcloud, PCL_POINTCLOUD, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(Position, position, POSITION, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(Orientation, orientation, ORIENTATION, NDArray, ndarray, NDARRAY) \
    RUN_ARON_MACRO(Pose, pose, POSE, NDArray, ndarray, NDARRAY)

#define HANDLE_ENUM_CORRESPONDING \
    RUN_ARON_MACRO(IntEnum, intenum, INT_ENUM, Int, int, INT)

#define HANDLE_PRIMITIVE_CORRESPONDING \
    RUN_ARON_MACRO(Int, int, INT, Int, int, INT) \
    RUN_ARON_MACRO(Long, long, LONG, Long, long, LONG) \
    RUN_ARON_MACRO(Float, float, FLOAT, Float, float, FLOAT) \
    RUN_ARON_MACRO(Double, double, DOUBLE, Double, double, DOUBLE) \
    RUN_ARON_MACRO(String, string, STRING, String, string, STRING) \
    RUN_ARON_MACRO(Bool, bool, BOOL, Bool, bool, BOOL) \
    RUN_ARON_MACRO(Time, time, TIME, Long, long, LONG)

#define HANDLE_ALL_CORRESPONDING \
    HANDLE_CONTAINER_CORRESPONDING \
    HANDLE_COMPLEX_CORRESPONDING \
    HANDLE_ENUM_CORRESPONDING \
    HANDLE_PRIMITIVE_CORRESPONDING

/************************************
 * Related data types ***************
 * (which on can be converted into what)
 * automatically symmetric but not transitive
 ***********************************/
#define HANDLE_ARON_PRIMITIVE_DATA_RELATED \
    RUN_ARON_MACRO(Int, int, INT, Long, long, LONG) \
    RUN_ARON_MACRO(Int, int, INT, Float, float, FLOAT) \
    RUN_ARON_MACRO(Int, int, INT, Double, double, DOUBLE) \
    RUN_ARON_MACRO(Long, long, LONG, Float, float, FLOAT) \
    RUN_ARON_MACRO(Long, long, LONG, Double, double, DOUBLE) \
    RUN_ARON_MACRO(Float, float, FLOAT, Double, double, DOUBLE)
