/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "OpenCVMat.h"

namespace armarx::aron::cppcodegenerator::serializer
{

    const std::map<std::string, std::pair<std::string, int>> OpenCVMatSerializer::ACCEPTED_TYPES =
    {
        {"CV_8U", {"CV_8U", 1}},
        {"CV_8S", {"CV_8S", 1}},
        {"CV_16U", {"CV_16U", 2}},
        {"CV_16S", {"CV_16S", 2}},
        {"CV_32S", {"CV_32S", 4}},
        {"CV_32F", {"CV_32F", 4}},
        {"CV_64F", {"CV_64F", 8}},
    };

    // constructors
    OpenCVMatSerializer::OpenCVMatSerializer(const typenavigator::OpenCVMatNavigatorPtr& e) :
        Serializer("cv::Mat", simox::meta::get_type_name(typeid(data::AronNDArray)), simox::meta::get_type_name(typeid(type::AronOpenCVMat))),
        typenavigator(e)
    {
        // check if type exists
        if (ACCEPTED_TYPES.find(e->getTypename()) == ACCEPTED_TYPES.end())
        {
            throw error::StringNotValidException("OpenCVMatSerializer", "OpenCVMatSerializer", "Could not find a matching typename.", e->getTypename(), e->getPath());
        }
    }

    std::vector<CppFieldPtr> OpenCVMatSerializer::getPublicVariableDeclarations(const std::string& name) const
    {
        CppFieldPtr field = CppFieldPtr(new CppField(getFullCppTypename(), name));
        return {field};
    }

    std::vector<std::pair<std::string, std::string>> OpenCVMatSerializer::getCtorInitializers(const std::string&) const
    {
        return {};
    }

    CppBlockPtr OpenCVMatSerializer::getCtorBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getResetBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_dimensions = escaped_accessor + DIMENSION_ACCESSOR;
        std::string accessor_iterator = escaped_accessor + ITERATOR_ACCESSOR;

        b->addLine("std::vector<int> " + accessor_dimensions + ";");
        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + ".dims; ++" + accessor_iterator + ")");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor_dimensions + ".push_back(" + accessor + ".size[" + accessor_iterator + "]);");
        b->appendBlock(b2);
        b->addLine(accessor + " = " + getFullCppTypename() + "(" + accessor_dimensions + ", " + accessor + ".type());");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getInitializeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine(accessor + " = " + getFullCppTypename() + "(std::vector<int>({" + simox::alg::to_string<int>(typenavigator->getDimensions(), ", ") + "}), " + ACCEPTED_TYPES.at(typenavigator->getTypename()).first + ");");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getWriteInitialTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeOpenCVMat({" + simox::alg::to_string<int>(typenavigator->getDimensions(), ", ") + "}, \"" + typenavigator->getTypename() + "\");");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getWriteCurrentTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_dimensions = escaped_accessor + DIMENSION_ACCESSOR;
        std::string accessor_iterator = escaped_accessor + ITERATOR_ACCESSOR;

        b->addLine("std::vector<int> " + accessor_dimensions + ";");
        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + ".dims; ++" + accessor_iterator + ")");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor_dimensions + ".push_back(" + accessor + ".size[" + accessor_iterator + "]);");
        b->appendBlock(b2);
        b->addLine("w.writeOpenCVMat(" + accessor_dimensions + ", std::to_string(" + accessor + ".type()));");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_dimensions = escaped_accessor + DIMENSION_ACCESSOR;
        std::string accessor_iterator = escaped_accessor + ITERATOR_ACCESSOR;

        b->addLine("std::vector<int> " + accessor_dimensions + ";");
        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + ".dims; ++" + accessor_iterator + ")");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor_dimensions + ".push_back(" + accessor + ".size[" + accessor_iterator + "]);");
        b->appendBlock(b2);
        b->addLine(accessor_dimensions + ".push_back(" + accessor + ".elemSize());");
        b->addLine("w.writeNDArray(" + accessor_dimensions + ", std::to_string(" + accessor + ".type()), reinterpret_cast<const unsigned char*>(" + accessor + ".data));");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_dimensions = escaped_accessor + DIMENSION_ACCESSOR;
        std::string accessor_type = escaped_accessor + TYPE_ACCESSOR;

        b->addLine("const auto [" + accessor_dimensions + ", " + accessor_type + "] = r.readStartNDArray();");
        b->addLine(accessor + " = " + getFullCppTypename() + "(std::vector<int>({" + accessor_dimensions + ".begin(), std::prev(" + accessor_dimensions + ".end())}), std::stoi(" + accessor_type + "));");
        b->addLine("r.readEndNDArray(reinterpret_cast<unsigned char*>(" + accessor + ".data));");
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("if (cv::countNonZero(" + accessor + " != " + otherInstanceAccessor + ") != 0)");
        b->addLine("\t return false;");
        return b;
    }
}

