/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "IntEnumClass.h"

namespace armarx::aron::cppcodegenerator::serializer
{
    // constructors
    IntEnumClassSerializer::IntEnumClassSerializer(const typenavigator::IntEnumNavigatorPtr& n) :
        Serializer(n->getEnumName(), simox::meta::get_type_name(typeid(data::AronNDArray)), simox::meta::get_type_name(typeid(type::AronIntEnum))),
        navigator(n),
        enumName("__ImplEnum")
    {
    }

    std::vector<CppFieldPtr> IntEnumClassSerializer::getPublicVariableDeclarations(const std::string&) const
    {
        std::vector<CppFieldPtr> fields;
        std::stringstream enum_to_name;
        std::stringstream name_to_enum;
        std::stringstream enum_to_value;
        std::stringstream value_to_enum;

        enum_to_name << "{" << std::endl;
        name_to_enum << "{" << std::endl;
        enum_to_value << "{" << std::endl;
        value_to_enum << "{" << std::endl;
        for (const auto& [key, value] : navigator->getAcceptedValues())
        {
            std::string enumKeyWithNamespace = enumName + "::" + key;
            fields.push_back(std::make_shared<CppField>("const static " + enumName, key + " = " + enumKeyWithNamespace));

            enum_to_name << "\t\t{" << enumKeyWithNamespace << ", \"" << key << "\"}," << std::endl;
            name_to_enum << "\t\t{\"" << key << "\", " << enumKeyWithNamespace << "}," << std::endl;

            enum_to_value << "\t\t{" << enumKeyWithNamespace << ", " << value << "}," << std::endl;
            value_to_enum << "\t\t{" << value << ", " << enumKeyWithNamespace << "}," << std::endl;
        }

        enum_to_name << "\t}";
        name_to_enum << "\t}";
        enum_to_value << "\t}";
        value_to_enum << "\t}";

        fields.push_back(std::make_shared<CppField>("const std::map<" + enumName + ", std::string>", "EnumToStringMap", enum_to_name.str()));
        fields.push_back(std::make_shared<CppField>("const std::map<std::string, " + enumName + ">", "StringToEnumMap", name_to_enum.str()));
        fields.push_back(std::make_shared<CppField>("const std::map<" + enumName + ", int>", "EnumToValueMap", enum_to_value.str()));
        fields.push_back(std::make_shared<CppField>("const std::map<int, " + enumName + ">", "ValueToEnumMap", value_to_enum.str()));

        fields.push_back(std::make_shared<CppField>(enumName, "value"));

        return fields;
    }

    std::vector<std::pair<std::string, std::string>> IntEnumClassSerializer::getCtorInitializers(const std::string&) const
    {
        return {};
    }

    CppBlockPtr IntEnumClassSerializer::getCtorBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getResetBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("value = {};");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getWriteInitialTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeInt();");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getWriteCurrentTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeInt();");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getInitializeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("value = {};");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getWriteBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writePrimitive(EnumToValueMap.at(value));");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getReadBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("int temporary;");
        b->addLine("r.readPrimitive(temporary);");
        b->addLine("value = ValueToEnumMap.at(temporary);");
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getEqualsBlock(const std::string&, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("if (not (value == " + otherInstanceAccessor + ".value))");
        b->addLine("\t return false;");
        return b;
    }


    CppCtorPtr IntEnumClassSerializer::toCopyCtor(const std::string& name) const
    {
        CppCtorPtr c = std::make_shared<CppCtor>(name + "(const " + getFullCppTypename() + "& i)");
        std::vector<std::pair<std::string, std::string>> initList = {{"value", "i.value"}};
        c->addInitListEntries(initList);
        return c;
    }

    CppCtorPtr IntEnumClassSerializer::toInnerEnumCtor(const std::string& name) const
    {
        CppCtorPtr c = std::make_shared<CppCtor>(name + "(const " + enumName + " e)");
        std::vector<std::pair<std::string, std::string>> initList = {{"value", "e"}};
        c->addInitListEntries(initList);
        return c;
    }

    CppEnumPtr IntEnumClassSerializer::toInnerEnumDefinition() const
    {
        CppEnumPtr e = std::make_shared<CppEnum>(enumName, "The internal enum definition of the enum of this autogenerated class.");
        for (const auto& [key, value] : navigator->getAcceptedValues())
        {
            e->addField(std::make_shared<CppEnumField>(key));
        }
        return e;
    }

    CppMethodPtr IntEnumClassSerializer::toIntMethod() const
    {
        std::stringstream doc;
        doc << "@brief int() - Converts the internally stored value to int representation \n";
        doc << "@return - the int representation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("operator int() const", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("return EnumToValueMap.at(value);");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toCopyAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for copy \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod(getFullCppTypename() + "& operator=(const " + getFullCppTypename() + "& c)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("value = c.value;");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toEnumAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for the internally defined enum \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod(getFullCppTypename() + "& operator=(" + enumName + " v)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("value = v;");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toIntAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for the internally defined enum \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void operator=(int v)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("if (auto it = ValueToEnumMap.find(v); it == ValueToEnumMap.end())");
        CppBlockPtr b2 = std::make_shared<CppBlock>();
        b2->addLine("throw armarx::LocalException(\"The input int is not valid. Could net set the enum to value '\" + std::to_string(v) + \"'\");");
        b->addBlock(b2);
        b->addLine("else");
        CppBlockPtr b3 = std::make_shared<CppBlock>();
        b3->addLine("value = it->second;");
        b->addBlock(b3);
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toToStringMethod() const
    {
        std::stringstream doc;
        doc << "@brief toString() - Converts the internally stored value to string \n";
        doc << "@return - the name of the enum";

        CppMethodPtr m = CppMethodPtr(new CppMethod("std::string toString() const", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("return EnumToStringMap.at(value);");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toFromStringMethod() const
    {
        std::stringstream doc;
        doc << "@brief fromString() - sets the internally stored value to the corrsponding enum of the input str \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void fromString(const std::string& str)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("if (auto it = StringToEnumMap.find(str); it == StringToEnumMap.end())");
        CppBlockPtr b2 = std::make_shared<CppBlock>();
        b2->addLine("throw armarx::LocalException(\"The input name is not valid. Could net set the enum to value '\" + str + \"'\");");
        b->addBlock(b2);
        b->addLine("else");
        CppBlockPtr b3 = std::make_shared<CppBlock>();
        b3->addLine("value = it->second;");
        b->addBlock(b3);

        m->setBlock(b);
        return m;
    }
}
