/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Converter.h"

#include <RobotAPI/libraries/aron/core/Debug.h>



namespace armarx::aron::dataIO
{
    void Converter::ReadAndConvert(ReaderInterface& reader, WriterInterface& writer, const aron::typenavigator::NavigatorPtr& expectedStructure)
    {
        type::Descriptor t_desc = expectedStructure != nullptr ? expectedStructure->getDescriptor() : type::Descriptor::eUnknown;
        //std::cout << "Expected Structure: " << type::DESCRIPTOR_TO_STRING(t_desc) << std::endl;
        data::Descriptor desc = reader.getTypeOfNext(t_desc);
        switch (desc)
        {
            case data::Descriptor::eDict:
            {
                int elements = reader.readStartDict();
                writer.writeStartDict();

                typenavigator::NavigatorPtr childType = nullptr;
                if (t_desc == type::Descriptor::eDict)
                {
                    auto t = typenavigator::DictNavigator::DynamicCastAndCheck(expectedStructure);
                    childType = t->getAcceptedType();
                }

                for (int i = 0; i < elements; ++i)
                {
                    std::string key = reader.readKey();
                    writer.writeKey(key);

                    if (t_desc == type::Descriptor::eObject)
                    {
                        auto t = typenavigator::ObjectNavigator::DynamicCastAndCheck(expectedStructure);
                        childType = t->getMemberType(key);
                    }

                    Converter::ReadAndConvert(reader, writer, childType);
                }

                writer.writeEndDict();
                reader.readEndDict();

                break;
            }
            case data::Descriptor::eList:
            {
                int elements = reader.readStartList();
                writer.writeStartList();

                typenavigator::NavigatorPtr childType = nullptr;
                if (t_desc == type::Descriptor::eList)
                {
                    auto t = typenavigator::ListNavigator::DynamicCastAndCheck(expectedStructure);
                    childType = t->getAcceptedType();
                }

                for (int i = 0; i < elements; ++i)
                {
                    if (t_desc == type::Descriptor::eObject)
                    {
                        auto t = typenavigator::TupleNavigator::DynamicCastAndCheck(expectedStructure);
                        childType = t->getAcceptedType(i);
                    }

                    Converter::ReadAndConvert(reader, writer, childType);
                }

                writer.writeEndList();
                reader.readEndList();
                break;
            }

            case data::Descriptor::eNDArray:
            {
                auto [dims, type] = reader.readStartNDArray();
                int elements = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
                std::vector<unsigned char> data(elements);
                reader.readEndNDArray(data.data());

                writer.writeNDArray(dims, type, data.data());
                break;
            }

            case data::Descriptor::eInt:
            {
                int val = reader.readPrimitive<int>();
                writer.writePrimitive(val);
                break;
            }
            case data::Descriptor::eLong:
            {
                long val = reader.readPrimitive<long>();
                writer.writePrimitive(val);
                break;
            }
            case data::Descriptor::eFloat:
            {
                float val = reader.readPrimitive<float>();
                writer.writePrimitive(val);
                break;
            }
            case data::Descriptor::eDouble:
            {
                double val = reader.readPrimitive<double>();
                writer.writePrimitive(val);
                break;
            }
            case data::Descriptor::eString:
            {
                std::string val = reader.readPrimitive<std::string>();
                writer.writePrimitive(val);
                break;
            }
            case data::Descriptor::eBool:
            {
                bool val = reader.readPrimitive<bool>();
                writer.writePrimitive(val);
                break;
            }

            default:
            {
                throw error::DescriptorNotValidException("Converter", "ReadAndConvert", "Data-Type could not be resolved", desc);
            }
        }
    }
}
