/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller (fabian dot peller at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NlohmannJSONWriter.h"

namespace armarx::aron::typeIO::writer
{

    void NlohmannJSONWriter::writeStartObject(const std::string& n)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_NAME_SLUG] = n;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eObject, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndObject()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartList()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eList, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartDict()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eDict, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartTuple()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eTuple, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndTuple()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartPair()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::ePair, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndPair()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeEigenMatrix(const std::vector<int>& dims, const std::string& t)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "EigenMatrix";
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = dims;
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeEigenQuaternion(const std::string& t)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "EigenQuaternion";
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeIVTCByteImage(unsigned int w, unsigned int h, const std::string& t)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "IVTCByteImage";
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = {w, h};
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeOpenCVMat(const std::vector<int>& dims, const std::string& t)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "OpenCVMat";
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = dims;
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePCLPointCloud(unsigned int w, unsigned int h, const std::string& t)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "PCLPointCloud";
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = {w, h};
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePosition()
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Position";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeOrientation()
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Orientation";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePose()
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Pose";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeInt()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_INT_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeLong()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_LONG_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeFloat()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_FLOAT_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeDouble()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_DOUBLE_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeString()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_STRING_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeBool()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_BOOL_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeTime()
    {
        auto token = stack.top();
        nlohmann::json j(io::Data::READER_WRITER_TIME_TYPENAME_SLUG);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }
}
