/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "PCLPointCloud.h"

namespace armarx::aron::typenavigator
{
    // constructors
    PCLPointCloudNavigator::PCLPointCloudNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::ePCLPointCloud, path),
        Navigator(type::Descriptor::ePCLPointCloud, path),
        type(new type::AronPCLPointCloud(0, 0, ""))
    {
    }

    PCLPointCloudNavigator::PCLPointCloudNavigator(const type::AronPCLPointCloudPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::ePCLPointCloud, path),
        Navigator(type::Descriptor::ePCLPointCloud, path),
        type(o)
    {
        CheckAronPtrForNull("AronPCLPointCloudTypeNavigator", "AronPCLPointCloudTypeNavigator", getPath(), o);
        checkTypename(type->typeName);
    }

    std::string PCLPointCloudNavigator::checkTypename(const std::string& s) const
    {
        for (const auto& [key, list] : ACCEPTED_TYPES)
        {
            if (s == key)
            {
                return key;
            }

            for (const auto& el : list)
            {
                if (s == el)
                {
                    return key;
                }
            }
        }
        throw error::AronException("PCLPointCloudNavigator", "checkTypename", "The typename is empty.", getPath());
    }

    type::AronPCLPointCloudPtr PCLPointCloudNavigator::toAronPCLPointCloudPtr() const
    {
        checkTypename(type->typeName);
        return type;
    }

    std::vector<int> PCLPointCloudNavigator::getDimensions() const
    {
        return {type->width, type->height};
    }

    unsigned int PCLPointCloudNavigator::getWidth() const
    {
        return type->width;
    }

    unsigned int PCLPointCloudNavigator::getHeight() const
    {
        return type->height;
    }

    std::string PCLPointCloudNavigator::getTypename() const
    {
        return type->typeName;
    }

    void PCLPointCloudNavigator::setWidth(const unsigned int& w)
    {
        if (w == 0)
        {
            throw error::AronException("AronPCLPointCloudTypeNavigator", "setWidth", "The width of aa pointcloud cannot be 0", getPath());
        }
        type->width = w;
    }

    void PCLPointCloudNavigator::setHeight(const unsigned int& h)
    {
        if (h == 0)
        {
            throw error::AronException("AronPCLPointCloudTypeNavigator", "setHeight", "The height of aa pointcloud cannot be 0", getPath());
        }
        type->height = h;
    }

    void PCLPointCloudNavigator::setTypename(const std::string& u)
    {
        type->typeName = checkTypename(u);
    }


    // static methods
    PCLPointCloudNavigatorPtr PCLPointCloudNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<PCLPointCloudNavigator>(n);
    }

    PCLPointCloudNavigatorPtr PCLPointCloudNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("AronPCLPointCloudTypeNavigator", "DynamicCast[Before]", n);
        PCLPointCloudNavigatorPtr casted = PCLPointCloudNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("AronPCLPointCloudTypeNavigator", "DynamicCast[After]", n->getPath(), casted);
        return casted;
    }

    // virtual implementations
    std::vector<NavigatorPtr> PCLPointCloudNavigator::getChildren() const
    {
        return {};
    }

    size_t PCLPointCloudNavigator::childrenSize() const
    {
        return 0;
    }

    type::AronTypePtr PCLPointCloudNavigator::getResult() const
    {
        return toAronPCLPointCloudPtr();
    }

    std::string PCLPointCloudNavigator::getName() const
    {
        return "AronPCLPointCloudType<" + simox::alg::to_string(getDimensions(), ", ") + ", " + type->typeName + ">";
    }
}

